@@ -26,7 +26,6 @@ import (
 	"github.com/miekg/dns"
 	"go.uber.org/zap"
 	"golang.org/x/net/http2"
-	"golang.org/x/net/proxy"
 	"io"
 	"net"
 	"net/http"
@@ -59,54 +58,45 @@ type Upstream interface {
 type Opt struct {
 	// DialAddr specifies the address the upstream will
 	// actually dial to.
+	// Not implemented for doh upstreams with http/3.
 	DialAddr string
 
-	Bootstrap string
-
 	// Socks5 specifies the socks5 proxy server that the upstream
-	// will connect though. Currently, only tcp, dot, doh upstream support Socks5 proxy.
+	// will connect though.
+	// Not implemented for udp upstreams and doh upstreams with http/3.
 	Socks5 string
 
-	// IdleTimeout used by tcp, dot, doh to control connection idle timeout.
-	// If negative, tcp, dot will not reuse connections.
-	// Default: tcp & dot: 10s , doh: 30s.
+	// SoMark specifies the mark for each packet sent through this upstream.
+	// Not implemented for doh upstreams with http/3
+	SoMark int
+
+	// IdleTimeout specifies the idle timeout for long-connections.
+	// Available for TCP, DoT, DoH.
+	// If negative, TCP, DoT will not reuse connections.
+	// Default: TCP, DoT: 10s , DoH: 30s.
 	IdleTimeout time.Duration
 
-	// EnablePipeline enables the query pipelining as RFC 7766 6.2.1.1 suggested.
-	// Available for tcp/dot upstream with IdleTimeout >= 0.
+	// EnablePipeline enables query pipelining support as RFC 7766 6.2.1.1 suggested.
+	// Available for TCP, DoT upstream with IdleTimeout >= 0.
 	EnablePipeline bool
 
-	// EnableHTTP3 enables the HTTP/3 for DoH. Note that there is no HTTP/2 fallback.
+	// EnableHTTP3 enables HTTP/3 protocol for DoH upstream.
 	EnableHTTP3 bool
 
 	// MaxConns limits the total number of connections, including connections
 	// in the dialing states.
-	// MaxConns takes effect on tcp/dot upstream with IdleTimeout >= 0 and EnablePipeline.
-	// And doh upstream.
+	// Implemented for TCP/DoT pipeline enabled upstreams and DoH upstreams.
 	// Default is 1.
 	MaxConns int
 
 	// TLSConfig specifies the tls.Config that the TLS client will use.
-	// Used by dot, doh.
+	// Available for DoT, DoH upstreams.
 	TLSConfig *tls.Config
 
 	// Logger specifies the logger that the upstream will use.
 	Logger *zap.Logger
 }
 
-func dialTCP(ctx context.Context, addr, socks5 string) (net.Conn, error) {
-	if len(socks5) > 0 {
-		socks5Dialer, err := proxy.SOCKS5("tcp", socks5, nil, nil)
-		if err != nil {
-			return nil, fmt.Errorf("failed to init socks5 dialer: %w", err)
-		}
-		return socks5Dialer.(proxy.ContextDialer).DialContext(ctx, "tcp", addr)
-	}
-
-	d := net.Dialer{}
-	return d.DialContext(ctx, "tcp", addr)
-}
-
 func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 	if opt == nil {
 		opt = new(Opt)
@@ -128,10 +118,7 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 			Logger: opt.Logger,
 			DialFunc: func(ctx context.Context) (net.Conn, error) {
 				d := net.Dialer{
-					Resolver: &net.Resolver{
-						PreferGo: true,
-						Dial:     nil,
-					},
+					Control: getSetMarkFunc(opt.SoMark),
 				}
 				return d.DialContext(ctx, "udp", dialAddr)
 			},
@@ -146,7 +133,9 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 		tt := &Transport{
 			Logger: opt.Logger,
 			DialFunc: func(ctx context.Context) (net.Conn, error) {
-				d := net.Dialer{}
+				d := net.Dialer{
+					Control: getSetMarkFunc(opt.SoMark),
+				}
 				return d.DialContext(ctx, "tcp", dialAddr)
 			},
 			WriteFunc: dnsutils.WriteMsgToTCP,
@@ -161,7 +150,7 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 		t := &Transport{
 			Logger: opt.Logger,
 			DialFunc: func(ctx context.Context) (net.Conn, error) {
-				return dialTCP(ctx, dialAddr, opt.Socks5)
+				return dialTCP(ctx, dialAddr, opt.Socks5, opt.SoMark)
 			},
 			WriteFunc:      dnsutils.WriteMsgToTCP,
 			ReadFunc:       dnsutils.ReadMsgFromTCP,
@@ -185,7 +174,7 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 		t := &Transport{
 			Logger: opt.Logger,
 			DialFunc: func(ctx context.Context) (net.Conn, error) {
-				conn, err := dialTCP(ctx, dialAddr, opt.Socks5)
+				conn, err := dialTCP(ctx, dialAddr, opt.Socks5, opt.SoMark)
 				if err != nil {
 					return nil, err
 				}
@@ -226,14 +215,11 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 					InitialConnectionReceiveWindow: 256 * 1024,
 					MaxConnectionReceiveWindow:     512 * 1024,
 				},
-				dialFunc: func(network, addr string, tlsCfg *tls.Config, cfg *quic.Config) (quic.EarlySession, error) {
-					return quic.DialAddrHostContext(context.Background(), dialAddr, addrURL.Host, tlsCfg, cfg, true)
-				},
 			}
 		} else {
 			t1 := &http.Transport{
 				DialContext: func(ctx context.Context, network, _ string) (net.Conn, error) { // overwrite server addr
-					return dialTCP(ctx, dialAddr, opt.Socks5)
+					return dialTCP(ctx, dialAddr, opt.Socks5, opt.SoMark)
 				},
 				TLSClientConfig:     opt.TLSConfig,
 				TLSHandshakeTimeout: tlsHandshakeTimeout,
@@ -254,7 +240,7 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 			t = t1
 		}
 
-		return &DoH{
+		return &DoHUpstream{
 			EndPoint: addr,
 			Client:   &http.Client{Transport: t},
 		}, nil
