@@ -22,6 +22,7 @@ import (
 	"github.com/lucas-clemente/quic-go"
 	"github.com/lucas-clemente/quic-go/http3"
 	"go.uber.org/zap"
+	"net"
 	"net/http"
 	"sync"
 	"time"
@@ -38,10 +39,10 @@ type h3rt struct {
 	logger     *zap.Logger
 	tlsConfig  *tls.Config
 	quicConfig *quic.Config
-	dialFunc   func(network, addr string, tlsCfg *tls.Config, cfg *quic.Config) (quic.EarlySession, error)
 
 	m  sync.Mutex
 	rt *http3.RoundTripper
+	c  net.PacketConn
 }
 
 func (h *h3rt) getLogger() *zap.Logger {
@@ -58,7 +59,6 @@ func (h *h3rt) getRT() *http3.RoundTripper {
 		h.rt = &http3.RoundTripper{
 			TLSClientConfig: h.tlsConfig,
 			QuicConfig:      h.quicConfig,
-			Dial:            h.dialFunc,
 		}
 	}
 	return h.rt
